from typing import Tuple

from .utils import (
    SafetyException,
    _calculate_M,
    _calculate_x,
    _generate_random_bytes,
    _get_srp_generator,
    _get_srp_prime,
    _Hash,
    _pad,
    _to_int,
)


class EvidenceException(Exception):
    """
    Exception raised when server evidence key does not match.
    """
    pass


def generate_a_pair() -> Tuple[int, int]:
    """
    Generate private ephemeral a and public key A.

    Returns:
        Tuple (private a (int), public A (int))
    """
    prime = _get_srp_prime()
    generator = _get_srp_generator()

    a = _to_int(_generate_random_bytes(32))  # RFC-5054 reccomends 256 bits

    A = pow(generator, a, prime)

    return a, A


def process_challenge(
    identity: str,
    password: str,
    salt: bytes,
    a: int,
    A: int,
    B: int,
) -> Tuple[bytes, bytes]:
    """
    Takes in salt and public value B to respond to SRP challenge with
    message M. Also returns session key for later authentication that
    the server is legit.

    Args:
        identity (str): the identity to process
        password (str): the password to process
        a        (int): the ephemeral value a generated by the client
        A        (int): the public value A generated by the client
        B        (int): the public value B from the server


    Returns:
        Tuple (message (bytes), session_key (bytes))

    Raises:
        SafetyException: if fails to pass SRP-6a safety checks
    """
    prime = _get_srp_prime()

    width = prime.bit_length()

    generator = _get_srp_generator()
    padded_generator = _pad(generator, width)

    padded_A = _pad(A, width)
    padded_B = _pad(B, width)

    # u - random scrambling param
    u = _to_int(_Hash(padded_A, padded_B))

    # x - private key
    x = _calculate_x(salt, identity, password)

    # k - multiplier
    k = _to_int(_Hash(prime, padded_generator))

    # SRP-6a safety checks
    if B == 0:
        raise SafetyException("Public value B is 0. Auth Failed.")

    if u == 0:
        raise SafetyException("Scrambler u is 0. Auth Failed.")

    # Premaster secret, S = (B - k*(generator^x)) ^ (a + u*x)
    t1 = B - k * pow(generator, x, prime)
    t2 = a + u * x

    # Calculate shared session key
    S = pow(t1, t2, prime)
    session_key = _Hash(S)

    # Shared message to server
    M = _calculate_M(
        generator,
        prime,
        identity,
        salt,
        A,
        B,
        session_key,
    )

    return M, session_key


def verify_session(
    A: int,
    M: bytes,
    session_key: bytes,
    server_H_AMK: bytes,
) -> bytes:
    """
    Verify session with server evidence key H_AMK.

    Args:
        A              (int): the public A value generated by the client
        M            (bytes): the message the client sends to the server
        session_key  (bytes): the strong private session key generated by the client
        server_H_AMK (bytes): the evidence key returned by the server
    """
    client_H_AMK = _Hash(A, M, session_key)

    if client_H_AMK != server_H_AMK:
        raise EvidenceException("Evidence keys do not match. Auth Failed.")

    return client_H_AMK



import hashlib
import os
from typing import Any

from .constants import SRP_GENERATOR, SRP_PRIME

BYTE_ORDER = "big"


class SafetyException(Exception):
    """
    Raised if SRP-6a safety checks fail
    """
    pass


def _get_srp_prime() -> int:

    return int(SRP_PRIME, 16)


def _get_srp_generator() -> int:

    return int(SRP_GENERATOR, 16)


def _generate_random_bytes(length: int) -> bytes:
    """
    Cryptographically random bytes
    """

    return os.urandom(length)


def _to_bytes(obj: Any) -> bytes:
    """
    Convert obj to bytes.

    Args:
        obj (int, str, bytes): the object to convert
    Returns:
        bytes representation of the object
    """
    if type(obj) == bytes:
        return obj
    elif type(obj) == int:
        return obj.to_bytes((obj.bit_length() + 7) // 8, byteorder=BYTE_ORDER)
    elif type(obj) == str:
        return bytes(obj, "utf-8")
    else:
        raise Exception("Unable to convert object to bytes.")


def _to_int(obj: Any) -> int:
    """
    Convert obj to integer.

    Args:
        obj (int, str, bytes): the object to convert
    Returns
        integer representation of the object
    """
    if type(obj) == int:
        return obj
    elif type(obj) == bytes:
        return int.from_bytes(obj, byteorder=BYTE_ORDER)
    elif type(obj) == str:
        return int(obj, 16)
    else:
        raise Exception("Unable to convert object to integer.")


def _pad(obj: Any, length: int) -> bytes:
    """
    Left pad byte-string representation of number.

    Args:
        obj (int, str, bytes): the object to pad
        length          (int): the bitlength to pad to
    Returns:
        padded        (bytes): the padded object
    """
    unpadded = _to_bytes(obj)

    return b'\x00' * ((length + 7) // 8 - len(unpadded)) + unpadded


def _Hash(*args) -> bytes:
    """
    Hash of concatenated argument objects.

    Args:
        arg  (int, str, bytes): args to hash
    Returns:
        hash           (bytes): the hash
    """
    hsh = hashlib.sha256()

    for arg in args:
        hsh.update(arg if type(arg) == bytes else _to_bytes(arg))
    return hsh.digest()


def _calculate_x(salt: bytes, identity: str, password: str) -> int:
    """
    Calculate the user secret parameter x.

    Args:
        salt     (bytes): the salt associated with the identity
        identity   (str): the identity
        password   (str): the password
    Returns:
        x          (int): the user secret parameter x
    """
    pre_salt = _Hash(identity, ":", password)
    post_salt = _Hash(salt, pre_salt)

    return int.from_bytes(post_salt, byteorder=BYTE_ORDER)


def _calculate_M(
    generator: int,
    prime: int,
    identity: str,
    salt: bytes,
    A: int,
    B: int,
    session_key: bytes,
) -> bytes:
    """
    Calculate evidence message M hash.

    Args:
        generator     (int): the generator of the prime
        prime         (int): the large safe prime
        identity      (str): the identity
        salt        (bytes): the salt
        A             (int): the public value A from the client
        B             (int): the public value B from the server
        session_key (bytes): the strong session key, individually computed
    Returns:
        message M   (bytes): the message shared by the client to the server
    """
    H_generator = _Hash(generator)
    H_prime = _Hash(prime)
    H_identity = _Hash(identity)
    H_xor = bytes(map(lambda i: i[0] ^ i[1], zip(H_generator, H_prime))) # noqa: W1636; pylint: disable=map-builtin-not-iterating
    return _Hash(H_xor, H_identity, salt, A, B, session_key)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bsrp",
    version="1.0.1",
    author="Abe Hoffman",
    author_email="abehoffman@me.com",
    license='MIT',
    description="B-first SRP-6a protocol implementation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/abehoffman/bsrp",
    install_requires=[],
    tests_require=[
        "pytest",
        "pytest-cov",
    ],
    package_dir={"": "lib"},
    packages=["bsrp"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Security',
    ],
    python_requires='>=3.5',
)



from typing import Tuple

from .utils import (
    SafetyException,
    _calculate_M,
    _calculate_x,
    _generate_random_bytes,
    _get_srp_generator,
    _get_srp_prime,
    _Hash,
    _pad,
    _to_int,
)


class MessageException(Exception):
    """
    Exception raised when the message from the client does not match.
    """
    pass


def generate_salt_and_verifier(identity: str, password: str) -> Tuple[bytes, int]:
    """
    Create salt and verification key from identity and password.

    Args:
        identity (str): the identity to generate with
        password (str): the password to generate with
    Returns:
        Tuple(salt (bytes), verifier (int))
    """

    salt = _generate_random_bytes(32)
    generator = _get_srp_generator()
    prime = _get_srp_prime()

    x = _calculate_x(salt, identity, password)

    verifier = pow(generator, x, prime)

    return salt, verifier


def generate_b_pair(verifier: int) -> Tuple[int, int]:
    """
    Generates private ephemeral b for later use and public
    value B for the client.

    Args:
        verifier (int): the verifer to generate the pair with
    Returns:
        Tuple(private b (int), public B (int))
    """
    prime = _get_srp_prime()

    width = prime.bit_length()

    generator = _get_srp_generator()
    padded_generator = _pad(generator, width)

    b = _to_int(_generate_random_bytes(32))  # RFC-5054 recommends 256 bits

    # k - multiplier
    k = _to_int(_Hash(prime, padded_generator))

    # B = (k*verifier + (generator^b) % prime) % prime
    B = (k * verifier + pow(generator, b, prime)) % prime

    return b, B


def verify_session(
        identity: str,
        salt: bytes,
        verifier: int,
        A: int,
        b: int,
        M_client: bytes,
    ) -> bytes:
    """
    After the client has sent public value A and message M,
    the server verifies the session and returns evidence key
    H_AMK back to the client for mutual authentication.

    Args:
        identity   (str): the identity
        salt     (bytes): the salt stored by the server
        verifier   (int): the verifier stored by the server
        A          (int): the public value A from the client
        b          (int): the private ephemeral b stored by the server
        M_client (bytes): the message from the client
    Returns:
        H_AMK    (bytes): the evidence key to prove server legitimacy
    Raises:
        SafetyException: if SRP-6a safety checks fail
        MessageException: if server computed message does not match client message (auth failed)
    """
    prime = _get_srp_prime()

    width = prime.bit_length()

    generator = _get_srp_generator()
    padded_generator = _pad(generator, width)

    # SRP-6a safety check
    if A % prime == 0:
        raise SafetyException("A mod prime is 0. Auth Failed.")

    # k - multiplier
    k = _to_int(_Hash(prime, padded_generator))

    # B = (k*verifier + (generator^b) % prime) % prime
    B = (k * verifier + pow(generator, b, prime)) % prime

    padded_A = _pad(A, width)
    padded_B = _pad(B, width)

    # Precalculate scrambler
    u = _to_int(_Hash(padded_A, padded_B))

    # Calculate shared session key
    # S = (A*(v^u))^b
    S = pow(A * pow(verifier, u, prime), b, prime)
    session_key = _Hash(S)

    # Calculate shared message
    M_server = _calculate_M(
        generator,
        prime,
        identity,
        salt,
        A,
        B,
        session_key
    )

    if M_client != M_server:
        raise MessageException("Messages do not match. Auth Failed.")

    H_AMK = _Hash(A, M_server, session_key)

    return H_AMK
